import inspect
import json
import os
import shutil
import sys
from pathlib import Path
from subprocess import check_call

import click
import polib

from oarepo_tools import validate_output_translations_dir, validate_source_paths

npm_proj_cwd = os.path.dirname(inspect.getfile(inspect.currentframe()))
npm_proj_env = dict(os.environ)


def ensure_i18next_output_translations(
    base_dir: Path, i18n_configuration: dict
) -> Path:
    """
    Checks if i18next output directory exists and contains the necessary `i18next.js` entrypoint.
    When missing, this function will createit.

    This will also install any necessary `react-i18next` dev dependencies from `package.json`.

    :param base_dir: Python package root directory (containing `setup.cfg` or `oarepo.yaml`)
    :param i18n_configuration:
    :return root directory path of i18next output translations
    :raises SystemExit
    """
    output_dir = validate_output_translations_dir(
        base_dir,
        i18n_configuration,
        "i18next_output_translations",
        create_if_missing=True,
    )
    if not output_dir:
        click.secho(
            f"configuration error: `i18next_output_translations` directory missing or invalid.",
            fg="red",
        )
        sys.exit(1)

    # check if i18next.js exists and if it does not, create it
    i18next_entrypoint = output_dir / "i18next.js"
    messages_dir = output_dir / "messages"
    if not messages_dir.exists():
        messages_dir.mkdir(parents=True)

    # check if messages/index.js exists and if it does not, create it
    messages_index = messages_dir / "index.js"
    if not messages_index.exists():
        messages_index.touch()

    if not i18next_entrypoint.exists():
        shutil.copy(Path(__file__).parent / "i18next.js", i18next_entrypoint)
        click.secho(f"Created i18next.js in {i18next_entrypoint}", fg="green")
    else:
        shutil.copy(Path(__file__).parent / "i18next.js", i18next_entrypoint)
        click.secho(f"Updated i18next.js in {i18next_entrypoint}", fg="yellow")

    # Make sure NPM project is installed & up-to-date
    click.secho("Installing / updating React-i18next dependencies", fg="green")
    check_call(
        ["npm", "install"],
        env=npm_proj_env,
        cwd=npm_proj_cwd,
    )

    for language in i18n_configuration.get("languages", ("cs", "en")):
        catalogue_dir = output_dir / "messages" / language / "LC_MESSAGES"
        if not catalogue_dir.exists():
            catalogue_dir.mkdir(parents=True)
            click.secho(f"Created {catalogue_dir}", fg="green")
        messages = catalogue_dir / "translations.json"
        if not messages.exists():
            messages.write_text(json.dumps({}))
            click.secho(f"Created {messages}", fg="green")

    return output_dir


def _json_to_pot(input_path: Path, output_path: Path) -> Path:
    # Generate gettext POT file from extracted JS translation strings
    check_call(
        ["npm", "run", "generate_pot", "--", input_path, output_path],
        env=npm_proj_env,
        cwd=npm_proj_cwd,
    )

    return output_path


def extract_i18next_messages(base_dir: Path, temp_dir: Path, i18n_configuration):
    """
    Extracts all JS(X) i18next translation keys from `i18next_source_paths`
    using `i18next-scanner` and stores it in a `messages.pot` catalogue in the root of `temp_dir`.

    :param base_dir: Python package root directory (containing `setup.cfg` or `oarepo.yaml`)
    :param temp_dir: a temporary directory to store results (to not to overwrite ones from babel)
    :param i18n_configuration:
    :return: path to the resulting `messages.pot` catalogue
    """

    i18next_source_paths = validate_source_paths(
        base_dir, i18n_configuration, "i18next_source_paths"
    )

    if not i18next_source_paths:
        click.secho(
            f"Skipping i18next extraction: no valid source paths",
            fg="yellow",
        )
        return

    npm_proj_env["LANGUAGES"] = ",".join(i18n_configuration["languages"] or ["en"])

    source_path_patterns = [
        os.path.join(str(source_path), "**/*.{js,jsx,ts,tsx}")
        for source_path in i18next_source_paths
    ]

    # Extract JS translations strings
    click.secho(
        f"Extracting i18next messages from sources matching {source_path_patterns} -> {temp_dir}",
        fg="green",
    )
    check_call(
        [
            "npm",
            "run",
            "extract_messages",
            "--",
            "--output",
            temp_dir,
            *source_path_patterns,
        ],
        env=npm_proj_env,
        cwd=npm_proj_cwd,
    )

    translations_file = temp_dir / "extracted-messages.json"
    extracted_data = json.loads(translations_file.read_text("utf-8"))

    # Fix any incorrectly extracted (e.g. by <Trans>) values, set all to ""
    for key in extracted_data.keys():
        extracted_data[key] = ""

    translations_file.write_text(json.dumps(extracted_data), "utf-8")

    messages_pot = temp_dir / "messages.pot"
    _json_to_pot(translations_file, messages_pot)

    # Cleanup helper json file
    translations_file.unlink()

    return messages_pot


def merge_i18next_messages_to_po(
    source_messages_file: Path, target_catalogue_file: Path
):
    """Merges messages from i18next formatted json with a target catalogue PO file entries.

    :param source_messages_file: path to a source i18next JSON messages file
    :param target_catalogue_file: path to a target catalogue PO file
    """
    source_messages = json.loads(source_messages_file.read_text("utf-8"))
    target_catalogue = polib.pofile(str(target_catalogue_file))

    target_catalogue_by_msgid = {entry.msgid: entry for entry in target_catalogue}

    for key, value in source_messages.items():
        if key in target_catalogue_by_msgid:
            if value:
                target_catalogue_by_msgid[key].msgstr = value
        else:
            target_catalogue.append(polib.POEntry(msgid=key, msgstr=value))

    target_catalogue.save(str(target_catalogue_file))


def merge_catalogues_from_i18next_translation_dir(
    source_translation_dir, target_translation_dir
):
    for source_catalogue_file in source_translation_dir.glob("*/translations.json"):
        click.secho(
            f"Merging i18next {source_catalogue_file} into {target_translation_dir}",
            fg="yellow",
        )
        language = source_catalogue_file.parent.name

        target_catalogue_file = (
            target_translation_dir / language / "LC_MESSAGES" / "messages.po"
        )
        if target_catalogue_file.exists():
            merge_i18next_messages_to_po(source_catalogue_file, target_catalogue_file)
        else:
            click.secho(
                f"Target catalogue file {target_catalogue_file} does not exist, "
                f"can not merge {source_catalogue_file}",
                fg="red",
            )


def compile_i18next_translations(
    source_translations_dir,
    output_translations_dir,
    i18n_configuration,
    skip_untranslated=True,
):
    """
    Compiles entries from source babel catalogue directory into
    i18next-compatible JSON format messages catalogue and updates
    messages module `index.js` to import all language-specific messages.

    :param source_translations_dir: path to a babel catalogue directory
    :param output_translations_dir: path to an i18next messages entrypoint directory
    :param i18n_configuration:
    """
    npm_proj_env["LANGUAGES"] = ",".join(i18n_configuration["languages"] or ["en"])

    click.secho(f"Compiling i18next messages in {source_translations_dir}", fg="green")
    check_call(
        [
            "npm",
            "run",
            "compile_catalog",
            "--",
            str(source_translations_dir),
            str(output_translations_dir),
            "--skip-untranslated" if skip_untranslated else "",
        ],
        env=npm_proj_env,
        cwd=npm_proj_cwd,
    )



import os
import shutil
from pathlib import Path

import polib

from oarepo_tools.babel import (
    compile_babel_translations,
    ensure_babel_configuration,
    ensure_babel_output_translations,
    extract_babel_messages,
    merge_babel_catalogues,
    merge_catalogue_dirs,
    update_babel_translations,
)
from tests.conftest import _clear_translations

jinjax_strings = ["jinjaxstring1"]
jinjax_extras = ["jinjaxstring2"]
python_strings = ["pythonstring1", "pythonstring2"]
html_strings = ["htmlstring1", "htmlstring2"]


def test_check_babel_configuration(app, db, cache, i18n_configuration, base_dir):
    babel_file = base_dir / "babel.ini"
    created_babel_file = ensure_babel_configuration(base_dir)

    assert str(created_babel_file) == str(babel_file)
    assert created_babel_file.exists()


def test_ensure_babel_output_translations(app, db, cache, i18n_configuration, base_dir):
    # Test create when missing
    config = i18n_configuration.copy()
    translations_dir = ensure_babel_output_translations(base_dir, config)
    assert translations_dir == base_dir / "mock_module/translations"
    # Check that files got created correctly
    paths = [
        translations_dir,
        translations_dir / "cs/LC_MESSAGES/messages.po",
        translations_dir / "en/LC_MESSAGES/messages.po",
        translations_dir / "da/LC_MESSAGES/messages.po",
    ]
    assert all([path.exists() for path in paths])

    # Test translations dir update
    config["languages"] = ["cs", "en", "da", "de"]
    translations_dir = ensure_babel_output_translations(base_dir, config)
    assert all(
        [
            path.exists()
            for path in paths + [translations_dir / "de/LC_MESSAGES/messages.po"]
        ]
    )

    # Test that with missing output config we bail out
    # TODO: need to use i18n_configuration copy here!!!
    del config["babel_output_translations"]
    try:
        translations_dir = ensure_babel_output_translations(base_dir, config)
    except SystemExit:
        pass


def test_extract_messages(
    app,
    db,
    cache,
    i18n_configuration,
    base_dir,
    babel_ini_file,
    babel_output_translations,
):
    messages_pot = extract_babel_messages(
        base_dir, babel_ini_file, babel_output_translations, i18n_configuration
    )

    assert messages_pot.exists()

    # Check if extra Jinjax strings got picked up
    # jinjax_messages = (tmpdir / "jinjax_messages.jinja").read_text(None)
    # assert all([f"{{{{ _('{js}') }}}}" in jinjax_messages for js in jinjax_extras])

    # Check all translation strings got extracted to POT file
    messages_catalogue = polib.pofile(str(messages_pot))
    entries = {entry.msgid: entry for entry in messages_catalogue}
    assert all(
        [
            key in entries.keys() and entries[key].msgstr == ""
            for key in jinjax_strings + jinjax_extras
        ]
    )
    assert all(
        [key in entries.keys() and entries[key].msgstr == "" for key in python_strings]
    )
    assert all(
        [key in entries.keys() and entries[key].msgstr == "" for key in html_strings]
    )

    # Ensure all translation strings are empty after extraction
    assert all([entry.msgstr == "" for entry in entries.values()])


def test_update_babel_translations(
    app,
    db,
    cache,
    i18n_configuration,
    base_dir,
    babel_ini_file,
    tmpdir,
    babel_output_translations,
):
    messages_pot = extract_babel_messages(
        base_dir, babel_ini_file, babel_output_translations, i18n_configuration
    )

    paths = [
        babel_output_translations / "cs/LC_MESSAGES/messages.po",
        babel_output_translations / "en/LC_MESSAGES/messages.po",
        babel_output_translations / "da/LC_MESSAGES/messages.po",
    ]
    assert all([os.path.getsize(str(path)) == 0 for path in paths])

    update_babel_translations(messages_pot, babel_output_translations)

    # Check all translation strings got propagated to language catalogs
    assert all([os.path.getsize(str(path)) > 0 for path in paths])
    for fpath in paths:
        po_file = polib.pofile(str(fpath))
        entries = {entry.msgid: entry for entry in po_file}

        assert all(
            [
                string in entries.keys()
                for string in jinjax_extras
                + jinjax_strings
                + html_strings
                + python_strings
            ]
        )

        # Check translations strings are initially empty
        assert all([entry.msgstr == "" for entry in entries.values()])


def test_compile_babel_translations(
    app,
    db,
    cache,
    i18n_configuration,
    babel_ini_file,
    tmpdir,
    base_dir,
    babel_output_translations,
):
    messages_pot = extract_babel_messages(
        base_dir, babel_ini_file, tmpdir, i18n_configuration
    )
    update_babel_translations(messages_pot, babel_output_translations)

    # Check that all .mo files got created
    compile_babel_translations(babel_output_translations)

    paths = [
        babel_output_translations / "cs/LC_MESSAGES/messages.mo",
        babel_output_translations / "en/LC_MESSAGES/messages.mo",
        babel_output_translations / "da/LC_MESSAGES/messages.mo",
    ]
    assert all([path.exists() and os.path.getsize(str(path)) > 0 for path in paths])


def test_merge_babel_catalogues(
    app, db, cache, i18n_configuration, base_dir, babel_output_translations, pofile
):
    source_path = babel_output_translations / "test_source.po"
    target_path = babel_output_translations / "test_target.po"

    # Simple addition
    source_entries = [
        polib.POEntry(
            msgid="Welcome",
            msgstr="Vitejte",
            occurrences=[("welcome.py", "12"), ("anotherfile.py", "34")],
        )
    ]
    target_entries = [
        polib.POEntry(
            msgid="Other",
            msgstr="Ostatni",
            occurrences=[("welcome.py", "2"), ("anotherfile.py", "3")],
        )
    ]

    pofile(source_entries, str(source_path))
    pofile(target_entries, str(target_path))

    merge_babel_catalogues(source_path, target_path)

    merged_catalogue = polib.pofile(target_path)
    merged_entries = {entry.msgid: entry for entry in merged_catalogue}
    assert all(
        [
            entry.msgid in merged_entries.keys()
            and merged_entries[entry.msgid].msgstr == entry.msgstr
            for entry in source_entries + target_entries
        ]
    )

    source_path.unlink()
    target_path.unlink()

    # Merge update
    source_entries = [
        polib.POEntry(
            msgid="Welcome",
            msgstr="Vitejte",
            occurrences=[("welcome.py", "12"), ("anotherfile.py", "34")],
        )
    ]
    target_entries = [
        polib.POEntry(
            msgid="Welcome",
            msgstr="Zdravim",
            occurrences=[("welcome.py", "12"), ("anotherfile.py", "34")],
        )
    ]
    pofile(source_entries, str(source_path))
    pofile(target_entries, str(target_path))

    merge_babel_catalogues(source_path, target_path)

    merged_catalogue = polib.pofile(target_path)
    merged_entries = {entry.msgid: entry for entry in merged_catalogue}
    assert all(
        [
            entry.msgid in merged_entries.keys()
            and merged_entries[entry.msgid].msgstr == entry.msgstr
            for entry in source_entries
        ]
    )

    source_path.unlink()
    target_path.unlink()
    target_path.with_suffix(".mo").unlink()


def test_merge_catalogue_dirs(
    app,
    db,
    cache,
    base_dir,
    babel_ini_file,
    babel_output_translations,
    i18n_configuration,
    extra_translations_dir,
):
    _clear_translations(i18n_configuration)
    messages_pot = extract_babel_messages(
        base_dir, babel_ini_file, babel_output_translations, i18n_configuration
    )
    update_babel_translations(messages_pot, babel_output_translations)

    target_translation_dir = ensure_babel_output_translations(
        extra_translations_dir, i18n_configuration
    )

    merge_catalogue_dirs(babel_output_translations, target_translation_dir)

    for catalogue_file in babel_output_translations.glob("*/LC_MESSAGES/*.po"):
        source_catalogue = polib.pofile(catalogue_file)
        merged_catalogue = polib.pofile(
            target_translation_dir
            / catalogue_file.relative_to(babel_output_translations)
        )
        source_entries = {entry.msgid: entry for entry in source_catalogue}
        merged_entries = {entry.msgid: entry for entry in merged_catalogue}
        assert all(
            [
                entry.msgid in merged_entries.keys()
                and merged_entries[entry.msgid].msgstr == entry.msgstr
                for entry in source_entries.values()
            ]
        )



import json
import os
import shutil
from pathlib import Path

import polib

from oarepo_tools.babel import (
    extract_babel_messages,
    merge_babel_catalogues,
    update_babel_translations,
)
from oarepo_tools.i18next import (
    compile_i18next_translations,
    ensure_i18next_output_translations,
    extract_i18next_messages,
    merge_catalogues_from_i18next_translation_dir,
    merge_i18next_messages_to_po,
)

jsstrings = ["jsstring1", "jsstring2"]
jinjax_strings = ["jinjaxstring1"]
jinjax_extras = ["jinjaxstring2"]
python_strings = ["pythonstring1", "pythonstring2"]
html_strings = ["htmlstring1", "htmlstring2"]


def test_ensure_i18next_output_translations(
    app, db, cache, base_dir, i18n_configuration
):
    config = i18n_configuration.copy()
    output_translations = ensure_i18next_output_translations(base_dir, config)

    paths = [
        output_translations,
        output_translations / "i18next.js",
        output_translations / "messages/index.js",
        output_translations / "messages/cs/LC_MESSAGES/translations.json",
        output_translations / "messages/en/LC_MESSAGES/translations.json",
        output_translations / "messages/da/LC_MESSAGES/translations.json",
    ]
    assert all([path.exists() for path in paths])

    del config["i18next_output_translations"]
    try:
        ensure_i18next_output_translations(base_dir, config)
    except SystemExit:
        pass


def test_extract_i18next_messages(app, db, cache, base_dir, tmpdir, i18n_configuration):
    messages_pot = extract_i18next_messages(base_dir, Path(tmpdir), i18n_configuration)

    # Check that files got created correctly
    assert messages_pot.exists()

    pot_catalogue = polib.POFile(messages_pot)
    pot_entries = {entry.msgid: entry for entry in pot_catalogue}
    assert all(
        [entry.msgid in jsstrings and entry.msgstr == "" for entry in pot_entries]
    )


def test_i18next_messages_to_po(app, db, cache, tmpdir, pofile):
    source_path = Path(tmpdir) / "test_source.json"
    target_path = Path(tmpdir) / "test_target.po"

    # Simple addition
    source_entries = {
        "Welcome": "Vitejte",
    }
    target_entries = [
        polib.POEntry(
            msgid="Other",
            msgstr="Ostatni",
            occurrences=[("welcome.py", "2"), ("anotherfile.py", "3")],
        )
    ]

    source_path.write_text(json.dumps(source_entries), "utf-8")
    pofile(target_entries, str(target_path))

    merge_i18next_messages_to_po(source_path, target_path)

    merged_catalogue = polib.pofile(target_path)
    merged_entries = {entry.msgid: entry for entry in merged_catalogue}
    assert all(
        [
            msgid in merged_entries.keys() and merged_entries[msgid].msgstr == msgstr
            for msgid, msgstr in [[k, v] for k, v in source_entries.items()]
            + [[e.msgid, e.msgstr] for e in target_entries]
        ]
    )

    source_path.unlink()
    target_path.unlink()

    # Merge update
    source_entries = {
        "Welcome": "Vitejte",
    }
    target_entries = [
        polib.POEntry(
            msgid="Welcome",
            msgstr="Zdravim",
            occurrences=[("welcome.py", "12"), ("anotherfile.py", "34")],
        )
    ]
    source_path.write_text(json.dumps(source_entries))
    pofile(target_entries, str(target_path))

    merge_i18next_messages_to_po(source_path, target_path)

    merged_catalogue = polib.pofile(target_path)
    merged_entries = {entry.msgid: entry for entry in merged_catalogue}
    assert all(
        [
            msgid in merged_entries.keys() and merged_entries[msgid].msgstr == msgstr
            for msgid, msgstr in [[k, v] for k, v in source_entries.items()]
        ]
    )

    source_path.unlink()
    target_path.unlink()


def test_merge_catalogues_from_i18next_translation_dir(
    app,
    db,
    cache,
    tmpdir,
    base_dir,
    babel_ini_file,
    babel_output_translations,
    i18n_configuration,
):
    i18next_messages_pot = extract_i18next_messages(
        base_dir, Path(tmpdir), i18n_configuration
    )
    babel_messages_pot = extract_babel_messages(
        base_dir, babel_ini_file, babel_output_translations, i18n_configuration
    )

    merge_babel_catalogues(i18next_messages_pot, babel_messages_pot)
    update_babel_translations(babel_messages_pot, babel_output_translations)

    source_translations_dir = base_dir / "test_i18next_translations"
    (source_translations_dir / "cs").mkdir(exist_ok=True, parents=True)
    (source_translations_dir / "en").mkdir(exist_ok=True, parents=True)
    (source_translations_dir / "da").mkdir(exist_ok=True, parents=True)
    (source_translations_dir / "en/translations.json").write_text(
        json.dumps({"jsstring2": "test", "jsstring3": ""})
    )
    (source_translations_dir / "cs/translations.json").write_text(
        json.dumps({"jsstring2": "test", "jsstring3": ""})
    )
    (source_translations_dir / "da/translations.json").write_text(
        json.dumps({"jsstring2": "test", "jsstring3": ""})
    )

    merge_catalogues_from_i18next_translation_dir(
        source_translations_dir, babel_output_translations
    )

    # Check that files got created correctly
    paths = [
        babel_output_translations / "cs/LC_MESSAGES/messages.po",
        babel_output_translations / "en/LC_MESSAGES/messages.po",
        babel_output_translations / "da/LC_MESSAGES/messages.po",
    ]
    assert all([path.exists() for path in paths])

    expected = {"jsstring1": "", "jsstring2": "test", "jsstring3": ""}

    shutil.rmtree(source_translations_dir, ignore_errors=True)

    # And has correct contents
    for fpath in paths:
        po_file = polib.pofile(str(fpath))
        entries = {entry.msgid: entry for entry in po_file}
        assert all(
            [
                key in entries.keys() and entries[key].msgstr == value
                for key, value in expected.items()
            ]
        )


def test_compile_i18next_translations(
    app,
    db,
    cache,
    base_dir,
    babel_output_translations,
    babel_ini_file,
    i18n_configuration,
    tmpdir,
):
    babel_messages_pot = extract_babel_messages(
        base_dir, babel_ini_file, babel_output_translations, i18n_configuration
    )
    update_babel_translations(babel_messages_pot, babel_output_translations)

    i18n_translations = ensure_i18next_output_translations(base_dir, i18n_configuration)

    paths = [
        i18n_translations / "messages/cs/LC_MESSAGES/translations.json",
        i18n_translations / "messages/en/LC_MESSAGES/translations.json",
        i18n_translations / "messages/da/LC_MESSAGES/translations.json",
    ]

    compile_i18next_translations(
        babel_output_translations,
        i18n_translations,
        i18n_configuration,
        skip_untranslated=True,
    )

    assert all([path.exists() for path in paths])

    for path in paths:
        json_translations = json.loads(path.read_text())
        assert not any([value == "" for value in json_translations.values()])

    compile_i18next_translations(
        babel_output_translations,
        i18n_translations,
        i18n_configuration,
        skip_untranslated=False,
    )
    for path in paths:
        json_translations = json.loads(path.read_text())
        assert all(
            [
                string in json_translations.keys()
                for string in jinjax_strings
                + jinjax_extras
                + html_strings
                + python_strings
            ]
        )



import os
from pathlib import Path

from oarepo_tools.make_translations import main


def test_cli_with_oarepo_yaml(app, db, cache, extra_entry_points, cli_runner):
    config_file = Path(__file__).parent
    try:
        cli_runner(main([str(config_file)]))
    except SystemExit as se:
        assert se.code == 0


def test_cli_with_setup_cfg(app, db, cache, extra_entry_points, cli_runner):
    config_file = Path(__file__).parent
    try:
        cli_runner(main([str(config_file)]))
    except SystemExit as se:
        assert se.code == 0


def test_cli_with_config_autodetect(app, db, cache, extra_entry_points, cli_runner):
    stored_cwd = os.getcwd()
    os.chdir(Path(__file__).parent)
    try:
        cli_runner(main([]))
    except SystemExit as se:
        assert se.code == 0
    finally:
        os.chdir(stored_cwd)


def test_cli_with_empty_setupcfg(app, db, cache, extra_entry_points, cli_runner):
    base_dir = Path(__file__).parent
    empty_config_file = base_dir / "empty_setup.cfg"
    real_config_file = base_dir / "setup.cfg"
    backup_config_file = base_dir / "setup.cfg.bak"

    real_config_file.rename(base_dir / backup_config_file)
    empty_config_file.rename(real_config_file)

    try:
        cli_runner(main([]))
    except SystemExit as se:
        assert se.code == 0
    finally:
        real_config_file.rename(empty_config_file)
        backup_config_file.rename(real_config_file)
